// -----------------------------------------------------------------------------
// eval_mnist_c.cpp
// 
// Developed by Hiroki Nakahara
// Tokyo Institute of Technology
// 30, Jan., 2017
// 
// ----------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// include coefficients
#include "param_cnn_float.h"

// include testbench
#include "mnist_dat.h"
#include "mnist_idx.h"

int tiny_cnn_mnist_kernel( 
    float mnist_dat[28][28]
);


// #######################################################################
// 
// TestBench(Top Function) for MNIST
// 
// #######################################################################
int main()
{
	// ########################################################
	// Definition for Variables
	// ########################################################
	int i, j, k, loop;
	int result;

	for( loop = 0; loop < 1000; loop+=10){
		// execute kernel
		result = tiny_cnn_mnist_kernel( mnist_dat[loop]);

		printf("loop=%4d index(test)=%d label=%d\n", 
            loop, result, mnist_idx[loop]);
	} // end of loop

	// enf of testbench ---------------------------------------
		
	return 1;
}

int tiny_cnn_mnist_kernel( 
	float mnist_dat[28][28]
)
{
//    printf ("-------------- MNIST --------------\n");

    // buffer memories
    float f1[1][28][28];
    float f2[1][24][24];
    float f3[1*12*12][1][1];
    float f4[10][1][1];

    // loop variables
    int i, j, k, x, y;

    // set mnist dataset ----------------------------------------------------------
    for( y = 0; y < 28; y++){
        for( x = 0; x < 28; x++){
            f1[0][y][x] = mnist_dat[y][x];

            if( f1[0][y][x] == 0.0){
                printf(".");
            } else {
                printf("@");
            }

        }
        printf("\n");
    }

    // loop control variables
    int stride, ksiz, n_ofeat, n_infeat, infeat_siz;
    int ofeat, oy, inf_y, ox, inf_x, infeat, ky, kx;
    float tmp, std, x_mu, norm_x;
    float tmp0, tmp1, tmp2, tmp3;
    float max_val, max_val1, max_val2;

    stride = 1;
    ksiz   = 1;

    // layer 1 (convolutional layer) ----------------------------------------------
    stride = 1; // ずらす量
    ksiz   = 5; // カーネルサイズ
    n_ofeat = 1; // 出力特徴マップ数
    n_infeat = 1; // 入力特徴マップ数
    infeat_siz = 28; // 入力特徴マップの1辺サイズ

    for( ofeat = 0; ofeat < n_ofeat; ofeat++){
        oy = 0;
        for( inf_y = 0; inf_y < infeat_siz - 4; inf_y += stride){
            ox = 0;
            for( inf_x = 0; inf_x < infeat_siz - 4; inf_x += stride){
                tmp = 0.0;
                for( infeat = 0; infeat < n_infeat; infeat++){
                    for( ky = 0; ky < ksiz; ky++){
                        for( kx = 0; kx < ksiz; kx++){
                            tmp = tmp + (f1[infeat][inf_y+ky][inf_x+kx] 
                                * conv1W[ofeat][infeat][ky][kx]);
                        }
                    }
                    tmp = tmp  + conv1b[ofeat];
                }

                // ReLU activation function
                if( tmp < 0.0)
                    tmp = 0.0;

                f2[ofeat][oy][ox] = tmp;

                ox = ox + 1;
            }
            oy = oy + 1;
        }
    }

    // layer 2 (max. pooling) -----------------------------------------------------
    stride = 2; // 2x2カーネルをプーリングするため, 2ニューロン毎にずらす
    ksiz   = 2;
    n_ofeat = 1;
    n_infeat = 1;
    infeat_siz = 24;
    for( ofeat = 0; ofeat < n_ofeat; ofeat++){
        oy = 0;
        for( inf_y = 0; inf_y < infeat_siz; inf_y += stride){
            ox = 0;
            for( inf_x = 0; inf_x < infeat_siz; inf_x += stride){
                tmp0 = f2[ofeat][inf_y+0][inf_x+0];
                tmp1 = f2[ofeat][inf_y+0][inf_x+1];
                tmp2 = f2[ofeat][inf_y+1][inf_x+0];
                tmp3 = f2[ofeat][inf_y+1][inf_x+1];

                if( tmp0 > tmp1){
                    max_val1 = tmp0;
                } else {
                    max_val1 = tmp1;
                }

                if( tmp2 > tmp3){
                    max_val2 = tmp2;
                } else {
                    max_val2 = tmp3;
                }

                if( max_val1 > max_val2){
                    max_val = max_val1;
                } else {
                    max_val = max_val2;
                }

                f3[ofeat*12*12+oy*12+ox][0][0] = max_val; // フル結合層用に1次元に展開

                ox = ox + 1;
            }
            oy = oy + 1;
        }
    }

    // layer 5 (Fully connection layer) -----------------------------------
    stride = 1;
    ksiz   = 1;
    n_ofeat = 10;
    n_infeat = 1*12*12; 
    infeat_siz = 1; // フル結合層の特徴マップは1ニューロンで構成
    for( ofeat = 0; ofeat < n_ofeat; ofeat++){
        tmp = 0.0;
        for( infeat = 0; infeat < n_infeat; infeat++){
            tmp = tmp + (f3[infeat][0][0] * l1W[ofeat][infeat]);
        }
        
        tmp = tmp + l1b[ofeat];

        f4[ofeat][0][0] = tmp;        

    }
    //printf("\n");
        
    // decision the next movement
    // note, the code does not perform a softmax operation!
    int loop;
    float max_val_result;
    int max_idx;

    max_idx = 0;
    max_val_result = -9999.9;

    // 最大の値を持つニューロンのインデックスを推論結果に用いる
    // 学習しないので, softmax関数の計算は行わない
    for( loop = 0; loop < 10; loop++){
        if( f4[loop][0][0] > max_val_result){
            max_val_result = f4[loop][0][0];
            max_idx = loop;
        }
    }

    return max_idx;
}

// ------------------------------------------------------------------
//                          END OF PROGRAM
// ------------------------------------------------------------------
