#!/usr/bin/env python
# -*- coding: utf-8 -*-

import chainer
import chainer.functions as F
import chainer.links as L

import chainer.functions as F

class MnistTinyCNN(chainer.Chain):

    """
    サンプル用の超小型CNN
    """

    def __init__(self, channel=1, c1=1, f1=1*12*12, filter_size1=5):
        super(MnistTinyCNN, self).__init__(
            conv1=L.Convolution2D(channel, c1, filter_size1),
            l1=L.Linear(f1, 10),
        )

    def __call__(self, x):
        # param x --- chainer.Variable of array
        x.data = x.data.reshape((len(x.data), 1, 28, 28))
        h = F.relu(self.conv1(x))
        h = F.max_pooling_2d(h, 2)
        y = self.l1(h)
        return y

class MnistCNN(chainer.Chain):

    """
    CNNの本来の記述, (5層AlexNetベースのTensorFlowチュートリアルの記述を参考に)
    """

    def __init__(self, channel=1, c1=4, c2=4, f1=144, \
                 f2=8, filter_size1=3, filter_size2=3):
        super(MnistCNN, self).__init__(
            conv1=L.Convolution2D(channel, c1, filter_size1),
            conv2=L.Convolution2D(c1, c2, filter_size2),
            l1=L.Linear(f1, f2),
            l2=L.Linear(f2, 10)
        )

    def __call__(self, x):
        # param x --- chainer.Variable of array
        x.data = x.data.reshape((len(x.data), 1, 28, 28))
        h = F.relu(self.conv1(x))
        h = F.max_pooling_2d(h, 2)
        h = F.relu(self.conv2(h))
        h = F.max_pooling_2d(h, 2)
        h = F.dropout(F.relu(self.l1(h)))
        y = self.l2(h)
        return y

class MnistMLP(chainer.Chain):

    """
    多層パーセプトロンの記述例 (前回の掲載したプログラム)
    """
    def __init__(self, n_in=784, n_units=100, n_out=10):
        super(MnistMLP, self).__init__(
            l1=L.Linear(n_in, n_units),
            l2=L.Linear(n_units, n_units),
            l3=L.Linear(n_units, n_out),
        )

    def __call__(self, x):
        h1 = F.relu(self.l1(x))
        h2 = F.relu(self.l2(h1))
        return self.l3(h2)
