// -----------------------------------------------------------------------------
// mnist_aocl.cl
// 
// Developed by Hiroki Nakahara
// Tokyo Institute of Technology
// 25, Jan., 2017
// 
// ----------------------------------------------------------------------------
// include coefficients
#include "param_cnn_aocl.h"

 // ACL kernel for adding two input vectors
__attribute__((num_simd_work_items(1)))
__attribute__((reqd_work_group_size(64,1,1)))
__kernel void mnist_aocl(__global const float * restrict x, 
                         __global const float * restrict y, 
                         __global int * restrict z)
{
    // get index of the work item
    int index = get_global_id(0);

    // buffer memories
	float ftmp[28*28];

    // loop variables
    int i, j, k, tx, ty;

    // set mnist dataset ----------------------------------------------------------
#pragma unroll 1
    for( ty = 0; ty < 28; ty++){
#pragma unroll 1
        for( tx = 0; tx < 28; tx++){
			ftmp[ty*28+tx] = x[(index*28*28)+ty*28+tx];			
        }
    }

    // loop control variables
    int stride, ksiz, n_ofeat, n_infeat, infeat_siz;
    int ofeat, oy, inf_y, ox, inf_x, infeat, ky, kx;
    float tmp, std, x_mu, norm_x;
    float tmp0, tmp1, tmp2, tmp3;
    float max_val, max_val1, max_val2;

    stride = 1;
    ksiz   = 1;

    // layer 1 (convolutional layer) ----------------------------------------------
    stride = 1; 
    ksiz   = 5; 
    n_ofeat = 1;
    n_infeat = 1; 
    infeat_siz = 28; 

#pragma unroll 1
    for( ofeat = 0; ofeat < n_ofeat; ofeat++){
        oy = 0;
        for( inf_y = 0; inf_y < infeat_siz - 4; inf_y += stride){
            ox = 0;
            for( inf_x = 0; inf_x < infeat_siz - 4; inf_x += stride){
                tmp = 0.0;
                for( infeat = 0; infeat < n_infeat; infeat++){
#pragma unroll 1
                    for( ky = 0; ky < ksiz; ky++){
#pragma unroll 1
                        for( kx = 0; kx < ksiz; kx++){
							tmp = tmp + (ftmp[(infeat)*28*28+(inf_y+ky)*28+(inf_x+kx)] 
                                * conv1W[ofeat][infeat][ky][kx]);
                        }
                    }
                    tmp = tmp  + conv1b[ofeat];
                }

                // ReLU activation function
                if( tmp < 0.0)
                    tmp = 0.0;

				ftmp[ofeat*24*24+oy*24+ox] = tmp;

                ox = ox + 1;
            }
            oy = oy + 1;
        }
    }

    // layer 2 (max. pooling) -----------------------------------------------------
    stride = 2;
    ksiz   = 2;
    n_ofeat = 1;
    n_infeat = 1;
    infeat_siz = 24;

#pragma unroll 1
    for( ofeat = 0; ofeat < n_ofeat; ofeat++){
        oy = 0;
        for( inf_y = 0; inf_y < infeat_siz; inf_y += stride){
            ox = 0;
            for( inf_x = 0; inf_x < infeat_siz; inf_x += stride){

                tmp0 = ftmp[(ofeat)*24*24+(inf_y+0)*24+(inf_x+0)];
                tmp1 = ftmp[(ofeat)*24*24+(inf_y+0)*24+(inf_x+1)];
                tmp2 = ftmp[(ofeat)*24*24+(inf_y+1)*24+(inf_x+0)];
                tmp3 = ftmp[(ofeat)*24*24+(inf_y+1)*24+(inf_x+1)];

                if( tmp0 > tmp1){
                    max_val1 = tmp0;
                } else {
                    max_val1 = tmp1;
                }

                if( tmp2 > tmp3){
                    max_val2 = tmp2;
                } else {
                    max_val2 = tmp3;
                }

                if( max_val1 > max_val2){
                    max_val = max_val1;
                } else {
                    max_val = max_val2;
                }

				ftmp[ofeat*12*12+oy*12+ox] = max_val;

                ox = ox + 1;
            }
            oy = oy + 1;
        }
    }

    // layer 5 (Fully connection layer) -----------------------------------
    stride = 1;
    ksiz   = 1;
    n_ofeat = 10;
    n_infeat = 1*12*12;
    infeat_siz = 1;
    for( ofeat = 0; ofeat < n_ofeat; ofeat++){
        tmp = 0.0;
        for( infeat = 0; infeat < n_infeat; infeat++){
			tmp = tmp + (ftmp[infeat] * l1W[ofeat][infeat]);
        }
        
        tmp = tmp + l1b[ofeat];

		ftmp[ofeat] = tmp;        

    }

    // decision the next movement
    // note, the code does not perform a softmax operation!
    int loop;
    float max_val_result;
    int max_idx;

    max_idx = 0;
    max_val_result = -9999.9;

#pragma unroll 1
    for( loop = 0; loop < 10; loop++){
        if( ftmp[loop] > max_val_result){
            max_val_result = ftmp[loop];
            max_idx = loop;
        }
    }

    z[index] = max_idx | (index << 3);
}

// ------------------------------------------------------------------
//                          END OF PROGRAM
// ------------------------------------------------------------------
